#include <stdio.h>

// result of computation is a square grid of integers with this
// many rows and columns of elements
#define DIM 800

struct cuComplex {
	float r;
	float i;
};

__device__ cuComplex cuComplexMake(double r, double i)
{
	cuComplex c;
	c.r = r;
	c.i = i;
	return c;
}

__device__ float cuComplexMagnitude( cuComplex c )
{
	return sqrt( c.r * c.r + c.i * c.i );
}

__device__ cuComplex cuComplexMultiply(cuComplex lhs, cuComplex rhs)
{
	return cuComplexMake(
		lhs.r * rhs.r - lhs.i * rhs.i,
		lhs.i * rhs.r + lhs.r * rhs.i
	);
}

__device__ cuComplex cuComplexAdd(cuComplex lhs, cuComplex rhs)
{
	return cuComplexMake( lhs.r + rhs.r, lhs.i + rhs.i );
}

__device__ int cuMandelIters( cuComplex c )
{
	cuComplex z;
	int iters = 0;

	z = cuComplexMake( 0.0f, 0.0f );
	while ( iters < 1000 ) {
		z = cuComplexMultiply( z, z );
		z = cuComplexAdd( z, c );
		if ( cuComplexMagnitude( z ) > 2.0f ) {
			break;
		}
		iters++;
	}

	return iters;
}

__global__ void kernel( int *data )
{
	int i = blockIdx.x;
	int j = blockIdx.y;

	float x = -2.0f + (((float)i / DIM) * 4.0f);
	float y = -2.0f + (((float)j / DIM) * 4.0f);

	int iters = cuMandelIters( cuComplexMake( x, y ) );

	data[j * DIM + i ] = iters;
}

int main(void)
{
	int *host_data, *dev_data;

	const size_t bufSize = (DIM*DIM) * sizeof(int);

	host_data = (int *) malloc( bufSize );
	cudaMalloc( (void**) &dev_data, bufSize );

	dim3 grid(DIM, DIM);
	kernel<<<grid, 1>>>( dev_data );

	cudaMemcpy( host_data, dev_data, bufSize, cudaMemcpyDeviceToHost );

	FILE *out = fopen("result.txt", "w");

	for (int j = 0; j < DIM; j++) {
		for (int i = 0; i < DIM; i++) {
			if (i > 0) {
				fprintf(out, " ");
			}
			fprintf(out, "%d", host_data[j*DIM + i]);
		}
		fprintf(out, "\n");
	}
	fclose(out);

	printf("Success?\n");

	return 0;
}

// vim: syntax=c
